/********************************************************************\
 * action.h -- API for handling user interface activities.          *
 * Copyright (C) 2002 Linas Vepstas <linas@linas.org>               *
 * http://dwi.sourceforge.net                                       *
 *                                                                  *
 * This library is free software; you can redistribute it and/or    *
 * modify it under the terms of the GNU Lesser General Public       *
 * License as published by the Free Software Foundation; either     *
 * version 2.1 of the License, or (at your option) any later version.
 *                                                                  *
 * This library is distributed in the hope that it will be useful,  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU Lesser General Public License for more details.              *
 *                                                                  *
 * You should have received a copy of the GNU Lesser General Public *
 * License along with this program; if not, contact:                *
 *                                                                  *
 * Free Software Foundation           Voice:  +1-617-542-5942       *
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652       *
 * Boston, MA  02111-1307,  USA       gnu@gnu.org                   *
\********************************************************************/

/*
 * FUNCTION:
 * API for handling user interface activities.  "Actions" typically
 * get run as a result of the GUI user clicking on a button or performing
 * some other user interface activity.  An action stores state information
 * about what should happen when the button-click/signal/etc. occurs.
 * 
 * More precisely, actions are triggered by GTK signals issued from GTK
 * widgets.  The action stores the template of an SQL query that will be
 * run when the signal is caught.  Along with this, it stores the widgets
 * that will be queried to fill in the blanks in the SQL statement.
 * 
 * The action also stores a list of reports that need to be refreshed 
 * when the SQL query is completed.
 *
 * Not all actions need to perform SQL queries.  Some (wtokey) cause
 * data to be copied from the user interface widgets to the key-value
 * pair tree.  Others (wtoobj) copy the data to GLib-2.0 GObjects.
 *
 * HISTORY:
 * Linas Vepstas March 2002
 */

#ifndef DUI_ACTION_H
#define DUI_ACTION_H

#include <glib.h>
#include <gtk/gtkobject.h>

#include "window.h"
#include "database.h"
#include "duifieldmap.h"

DuiAction * dui_action_new (DuiWindow *parent, const char *name);
void dui_action_destroy (DuiAction *);

const char * dui_action_get_name (DuiAction *);

/* Set the database connection that will be used to perform queries on */
void dui_action_set_database (DuiAction * act, DuiDatabase *db);
DuiDatabase * dui_action_get_database (DuiAction * act);


/* Specify a signal that will invoke this action. 
 * The widgetname must specify a GtkWidget on which
 * there exists a signal of name 'signalname'.
 * The widgetname must specify the name of a widget 
 * that appears in the glade file for the window that 
 * this action is a part of.
 */

void dui_action_add_signal (DuiAction *act, const char *widgettname, const char *signalname);

/* Specify the sql table that will be queried in this action.
 * querytype must be one of 'select', 'update', 'insert', 'delete', 
 * 'tables' or 'columns'.
 * The 'tables' field must return the name of a valid SQL table.
 */
void dui_action_set_table (DuiAction *, DuiField * tables, 
                                        const char * querytype);

/** Add a term, which may be an sql select, sql update, sql where matcher,
 * or it might be a non-sql map */
void dui_action_add_term (DuiAction * act, DuiFieldMap *fm);

/* Cross-reference to the windows that will show the results.
 * "row_name" is used to target multiple results to one report,
 * which is (for example) typically a columned tree. (GtkCTree)
 */
void dui_action_set_report (DuiAction *, const char * name, 
                            const char * row_name);

/* Add a window that needs to be refreshed when this action is performed */
void dui_action_add_refresh (DuiAction *, const char * name);

/* Rerun the last query (and display it in a report).
 * this routine is used to refresh a window.
 */
void dui_action_rerun_last_query (DuiAction *act);


/* Add a chained action */
void dui_action_add_chain (DuiAction *act, const char * actionname);

void dui_action_do_realize (DuiAction *);
int dui_action_is_realized (DuiAction *);

/* Force the db to connect. This will cause any connect
 * errors to be reported.
 */
void dui_action_db_connect (DuiAction *act);
		  

/* Cause the indicated action to be run. */
int dui_action_run (DuiAction *act);
		  

#endif /* DUI_ACTION_H */

/* ========================== END OF FILE ============================ */
