/*
 * FILE:
 * duitxnquery.c
 *
 * FUNCTION:
 * Handler for database queries.
 *
 * HISTORY:
 * Linas Vepstas March 2002
 */

#include "config.h"

#include <string.h>

#include <glib.h>

#include "builder.h"
#include "database.h"
#include "dui-initdb.h"
#include "duifield.h"
#include "duifieldmap.h"
#include "duifilter.h"
#include "duiresolver.h"
#include "duitxnquery.h"
#include "perr.h"

typedef enum {
	QUERY_NONE,
	QUERY_SIMPLE,
	QUERY_TABLES,
	QUERY_FIELDS
} MetaQueryType;

/* Implements connection between the database and the fields
 * that acess the database */
struct DuiTxnQuery_s
{
	/* Special DB queries that don't use normal SQL channels */
	MetaQueryType meta_qtype;
	
	/* Table(s) to query */
	DuiField *table;

	/* The query builder */
	SqlBuilder *sql_builder;

	/* A portion of the sql query */
	SqlBuilderQType qtype;

	GList *terms;

	/* Cache of most recent sql query string; used for rerunning the query. */
	char * query_string;

	/* Database connection on which to send the query */
	DuiDatabase *database;
	DuiDBConnection *db_conn;

	/* name resolution infrastructure */
	DuiResolver * resolver;
};

/* =================================================================== */
/* XXX remove these, these should be in duifield-sql.c only */

/* Data is compared to a database field. */
struct where_field_s
{
	char * compareop;
};

#define WHERE(f) ((struct where_field_s *)(&((f)->u.priv)))

/* =================================================================== */

static void
dui_txnquery_init (DuiTxnQuery *q)
{
	q->meta_qtype = QUERY_NONE;
	q->table = NULL;
	q->sql_builder = NULL;

	q->qtype = 0;
	q->terms = NULL;

	q->query_string = NULL;
	q->database = NULL;
	q->db_conn = NULL;

	q->resolver = NULL;
}

DuiTxnQuery *
dui_txnquery_new (void)
{
	DuiTxnQuery *qry;
	qry = g_new0 (DuiTxnQuery, 1);
	dui_txnquery_init (qry);
	return qry;
}

void
dui_txnquery_destroy (DuiTxnQuery *q)
{
	GList *node;
	
	q->meta_qtype = QUERY_NONE;
	q->table = NULL;

	q->database = NULL;
	q->db_conn = NULL;

	/* remove the insert/update terms */
	for (node=q->terms; node; node=node->next)
	{
		DuiFieldMap *fm = node->data;
		dui_field_map_destroy (fm);
	}
	g_list_free (q->terms);

	sql_builder_destroy(q->sql_builder);
	q->sql_builder = NULL;
	q->qtype = 0;

	g_free (q);
}

/* =================================================================== */

void 
dui_txnquery_set_resolver (DuiTxnQuery * q, DuiResolver *res)
{
	if (!q) return;
	q->resolver = res;
}
	
void 
dui_txnquery_set_database (DuiTxnQuery * q, DuiDatabase *db)
{
	if (!q) return;
	q->database = db;
}
	
DuiDatabase *
dui_txnquery_get_database (DuiTxnQuery * q)
{
	if (!q) return NULL;
	return q->database;
}

/* =================================================================== */

void 
dui_txnquery_set_table (DuiTxnQuery *qry, DuiField * tabfld, 
                                      const char * sql_querytype)
{
	const char * tables;

	ENTER ("(qry=%p, qry=%s)", qry, sql_querytype);
	if (!qry || !sql_querytype || !tabfld) return;
	if (!strcasecmp (sql_querytype, "tables"))
	{
		qry->meta_qtype = QUERY_TABLES;
		qry->table = NULL;
		return;
	}

	if (!strcasecmp (sql_querytype, "fields"))
	{
		qry->meta_qtype = QUERY_FIELDS;
		qry->table = tabfld;
		return;
	}

	/* Set up the sql statement builder so that the next set of
	 * of add_term calls set up a valid sql statement. At this point,
	 * a table must be specified.
	 */
	
	if (!tabfld->get_field_value) return;
	tables = tabfld->get_field_value(tabfld);
	if (!tables) return;

	qry->meta_qtype = QUERY_SIMPLE;
	if (!strcasecmp (sql_querytype, "select")) qry->qtype = SQL_SELECT;
	else if (!strcasecmp (sql_querytype, "insert")) qry->qtype = SQL_INSERT;
	else if (!strcasecmp (sql_querytype, "update")) qry->qtype = SQL_UPDATE;
	else if (!strcasecmp (sql_querytype, "delete")) qry->qtype = SQL_DELETE;
	else 
	{
		SYNTAX ("unknown query type \'%s\'", sql_querytype);
		return;
	}

	qry->sql_builder = sql_builder_new ();
	sql_builder_table (qry->sql_builder, tables, qry->qtype);
}

/* =================================================================== */

void 
dui_txnquery_add_term (DuiTxnQuery *qry, DuiFieldMap *fm)
{
	if (!qry || !fm) return;

	dui_resolver_add_field (qry->resolver, &fm->source);
	dui_resolver_add_field (qry->resolver, &fm->target);

	/* If its a SELECT term then we can handle it statically.
	 * The name of the SELECT field will be stored as a const target
	 */
	if (DUI_FIELD_IS_TYPE(&fm->target,DUI_FIELD_CONST))
	{
		sql_builder_set_str (qry->sql_builder, fm->target.u.value, NULL);
	}
	else
	{
		qry->terms = g_list_append (qry->terms, fm);
	}
}

/* =================================================================== */

void
dui_txnquery_do_realize (DuiTxnQuery *qry)
{
	GList *node;
	for (node=qry->terms; node; node=node->next)
	{
		DuiFieldMap *fm = node->data;
		dui_field_map_resolve (fm);
	}
}

/* =================================================================== */

void 
dui_txnquery_connect (DuiTxnQuery *qry)
{
	if (!qry) return;
	qry->db_conn = dui_database_do_realize (qry->database);
}

/* =================================================================== */

DuiDBRecordSet *
dui_txnquery_run (DuiTxnQuery *qry)
{
	DuiDBRecordSet *recs = NULL;
	if (!qry) return NULL;

	switch (qry->meta_qtype)
	{
		case QUERY_NONE: break;
		case QUERY_SIMPLE: break;
		case QUERY_TABLES:
			if (qry->db_conn)
			{
				recs = dui_connection_tables(qry->db_conn);
				return recs;
			}
			break;
		case QUERY_FIELDS:
			if ((qry->db_conn) && (qry->table->get_field_value))
			{
				const char * table;
				table = qry->table->get_field_value(qry->table);
				PINFO ("perform fields query with table=\'%s\'", table);
				recs = dui_connection_table_columns(qry->db_conn, table);
				return recs;
			}
			break;
	}

	/* If we have data we need to query, then do the query.  */
	if (qry->sql_builder)
	{
		GList *node;
		SqlBuilder *dupe;
		short have_where_terms = 0;

		g_free (qry->query_string);
		qry->query_string = NULL;

		/* Make a copy of the query, which might be holdig statically
		 * compiled 'SELECT term FROM table' in it. We append dynamic
		 * WHERE terms below.  Note also: if the sql_builder is empty,
		 * then there MUST be some update terms; alternately, if the 
		 * sql_builder is not empty, there must NOT be any update terms,
		 * else bad SQL will be generated. */
		dupe = sql_builder_copy (qry->sql_builder);
	
		/* Loop, looking for UPDATE terms */
		for (node = qry->terms; node; node=node->next)
		{
			DuiFieldMap *fm = node->data;
			const char * fieldval;
			if (!DUI_FIELD_IS_TYPE(&fm->target,DUI_FIELD_SQL)) continue;
			
			fieldval = dui_field_map_get_value (fm);
			sql_builder_set_str (dupe, fm->target.fieldname, fieldval);
		}
	
		/* Loop again, this time for where terms */
		for (node = qry->terms; node; node=node->next)
		{
			DuiFieldMap *fm = node->data;
			const char * fieldval;
			
			if (!DUI_FIELD_IS_TYPE(&fm->target,DUI_FIELD_WHERE)) continue;
	
			fieldval = dui_field_map_get_value (fm);
			/* fields left blank don't match */
			fieldval = whitespace_filter (fieldval);
			if (fieldval)
			{
				sql_builder_where_str (dupe,
				   fm->target.fieldname, 
					fieldval, WHERE(&fm->target)->compareop);
				have_where_terms = 1;
			}
		}
	
		if (NULL == qry->db_conn)
		{
			/* now get the database too */
			qry->db_conn = dui_database_do_realize (qry->database);
		}

		/* Perform the query only if its not a wild-card
		 * update or delete.  Wildcard updates/deletes are baaad. 
		 * Basically, by doing this, we are trying to prevent
		 * stupid DUI XML file errors from doing a global clobber.
		 */
		if (have_where_terms || ((SQL_UPDATE != qry->qtype) && 
		                         (SQL_DELETE != qry->qtype)))
		{
			const char * stmt;

			/* Get the sql statement string */
			stmt = sql_builder_query (dupe);
			qry->query_string = g_strdup (stmt);
	
			/* Now run the query */
			if (qry->db_conn)
			{
				recs = dui_connection_exec(qry->db_conn, stmt);
			}
		}	
		sql_builder_destroy (dupe);
	}

	return recs;
}

/* =================================================================== */

DuiDBRecordSet *
dui_txnquery_rerun_last_query (DuiTxnQuery *qry)
{
	DuiDBRecordSet *recs = NULL;

	if (!qry) return NULL;
	if (!qry->db_conn || !qry->query_string) return NULL;

	recs = dui_connection_exec (qry->db_conn, qry->query_string);
	return recs;
}

/* ========================== END OF FILE ============================ */
