/********************************************************************\
 * report.c --  Fill in widget values based on database query       *
 * Copyright (C) 2002 Linas Vepstas <linas@linas.org>               *
 * http://dwi.sourceforge.net                                       *
 *                                                                  *
 * This library is free software; you can redistribute it and/or    *
 * modify it under the terms of the GNU Lesser General Public       *
 * License as published by the Free Software Foundation; either     *
 * version 2.1 of the License, or (at your option) any later version.
 *                                                                  *
 * This library is distributed in the hope that it will be useful,  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU Lesser General Public License for more details.              *
 *                                                                  *
 * You should have received a copy of the GNU Lesser General Public *
 * License along with this program; if not, contact:                *
 *                                                                  *
 * Free Software Foundation           Voice:  +1-617-542-5942       *
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652       *
 * Boston, MA  02111-1307,  USA       gnu@gnu.org                   *
\********************************************************************/
/*
 * FUNCTION:
 * Fill in widget values based on results of database query
 *
 * HISTORY:
 * Linas Vepstas March 2002
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gnome.h>

#include "perr.h"
#include "util.h"

#include "action.h"
#include "dui-initdb.h"
#include "duifield.h"
#include "duifield-gobj.h"
#include "duifield-gtk.h"
#include "duifield-hash.h"
#include "duifield-sql.h"
#include "duifieldmap.h"
#include "duifilter.h"
#include "duiresolver.h"
#include "duitxnreport.h"
#include "report.h"
#include "window.h"

struct DuiReport_s
{
	DuiWindow *window;       /* parent window */
	char * name;

	GList *rows;             /* list of nested/tree row iterators */
	DuiTxnReport *curr_row;

	/* list of actions that last drew into this window */
	GList *last_queries;

	DuiResolver *resolver;
};

/* ============================================================ */

DuiReport * 
dui_report_new (DuiWindow *win, const char * name)
{
	DuiReport * rpt;

	if (!win) return NULL;

	rpt = g_new (DuiReport, 1);

	rpt->window = win;
	rpt->name = g_strdup (name);
	rpt->rows = NULL;
	rpt->curr_row = NULL;
	rpt->last_queries = NULL;

	rpt->resolver = dui_resolver_new ();
	dui_window_add_report (win, rpt);
	return rpt;
}

void 
dui_report_destroy (DuiReport *rpt)
{
	GList *rnode;
	if (NULL == rpt) return;

	rpt->window = NULL;
	g_free(rpt->name);
	g_list_free (rpt->last_queries);

	for (rnode=rpt->rows; rnode; rnode=rnode->next)
	{
		DuiTxnReport *row = rnode->data;
		dui_txnreport_destroy (row);
	}
	g_list_free (rpt->rows);
	rpt->rows = NULL;

	dui_resolver_destroy (rpt->resolver);
	g_free (rpt);
}

/* ============================================================ */

static DuiTxnReport *
find_row_by_name (DuiReport *rpt, const char * wanted_name)
{
	GList *node;
	
	/* locate the target row.  */
	for (node=rpt->rows; node; node=node->next)
	{
		DuiTxnReport * row = node->data;
		const char * row_name = dui_txnreport_get_name(row);

		if (NULL==wanted_name && NULL==row_name) return row;
		if (NULL==wanted_name && 0==row_name[0]) return row;
		if (0==wanted_name[0] && NULL==row_name) return row;
		if (NULL==wanted_name) continue;
		if (NULL==row_name) continue;
		if (!strcmp (row_name, wanted_name)) return row;
	}

	return NULL;
}

void
dui_report_set_last_action (DuiReport *rpt, DuiAction *act, 
                            const char * row_name)
{
	DuiTxnReport *row;
	if (!rpt) return;

	row = find_row_by_name (rpt, row_name);
		
	if (!row || !act || 0 == dui_txnreport_get_nest(row))
	{
		g_list_free (rpt->last_queries);
		rpt->last_queries = NULL;
	}

	if (!act) return;

	rpt->last_queries = g_list_append (rpt->last_queries,  act);
}

/* ============================================================ */

void
dui_report_add_row (DuiReport *rpt, const char * name, int nest)
{
	DuiTxnReport *row;
	if (!rpt) return;
	
	row = dui_txnreport_new (name, nest);
	rpt->rows = g_list_append (rpt->rows, row);
	dui_txnreport_set_resolver (row, rpt->resolver);

	rpt->curr_row = row;
}

/* ============================================================ */

void
dui_report_add_term (DuiReport *rpt, DuiFieldMap *fm)
{
	if (!rpt || !fm) return;

	/* Its valid to have reports without a row iterator. However, 
	 * column definitions are stored with a row iterator, so ...
	 */
	if (NULL == rpt->curr_row)
	{
		dui_report_add_row (rpt, NULL, 0);
	}
	dui_txnreport_add_term (rpt->curr_row, fm);
}

/* ============================================================ */

int
dui_report_show_data (DuiReport *rpt, DuiDBRecordSet *recs, 
                      const char * row_name)
{
	DuiTxnReport *row = NULL;
	
	if (!rpt) return 0;

	ENTER ("(rpt=%p \'%s\', recs=%p, row=%s)", rpt, dui_report_get_name (rpt), recs, row_name);

	/* locate the target row.  */
	row = find_row_by_name (rpt, row_name);

	if (NULL == row)
	{
		PERR ("could not find the target row %s", row_name);
		return 0;
	}
	
	/* Check to see if there's any data to display at all.
	 * If there's not, then we don't realize the report widget, 
	 * we don't show anytihing, we bail from here, and let the
	 * next action in the chain take over. */

	if (dui_txnreport_is_empty (row, recs)) return 0;
	
	/* Make widget show up on screen & other intialization */
	dui_window_realize (rpt->window);
	/* dui_txnreport_realize (row); */
	dui_txnreport_run (row, recs);

	LEAVE ("(rpt=%p \'%s\', recs=%p, row=%s)", rpt, dui_report_get_name (rpt), recs, row_name);
	
	return 1;
}

/* ============================================================ */

const char * 
dui_report_get_name (DuiReport *rpt)
{
	if (!rpt) return NULL;
	if (rpt->name) return (rpt->name);
	return dui_window_get_name (rpt->window);
}

/* ============================================================ */

void
dui_report_do_realize (DuiReport *rpt)
{
	DuiInterface *dui;
	if (!rpt) return;
	ENTER ("(rpt=%p \'%s\')", rpt, dui_report_get_name (rpt));

	/* Convert widget names into pointers to actual widgets */
	dui_resolver_resolve_widgets (rpt->resolver, rpt->window);

	dui = dui_window_get_interface (rpt->window);
	dui_resolver_resolve_gobj (rpt->resolver, dui);
	dui_resolver_resolve_hash (rpt->resolver, dui);

	LEAVE ("(rpt=%p \'%s\')", rpt, dui_report_get_name (rpt));
}

/* ============================================================ */

void
dui_report_refresh (DuiReport *rpt)
{
	GList *node;
	if (!rpt) return;

	ENTER ("(rpt=%p \'%s\')", rpt, dui_report_get_name (rpt));
	/* no refresh needed if the main window widget doesn't exist */
	if (0 == dui_window_is_realized (rpt->window)) return;

	for (node=rpt->last_queries; node; node=node->next)
	{
		DuiAction *last_action = node->data;
		dui_action_rerun_last_query (last_action);
	}
	LEAVE ("(rpt=%p \'%s\')", rpt, dui_report_get_name (rpt));
}

/* ============================== END OF FILE =================== */
